from pathlib import Path
from shutil import copyfile

import click

from autoalbument.utils.templates import AutoAlbumentTemplate


@click.command()
@click.option(
    "--config-dir",
    type=click.Path(),
    required=True,
    help="Path to a directory where AutoAlbument will place config files.",
)
@click.option(
    "--task",
    type=click.Choice(["classification", "semantic_segmentation"]),
    required=True,
    help="Deep learning task (either classification or semantic segmentation)",
)
@click.option("--num-classes", type=int, required=True, help="Number of classes in the dataset.")
def main(config_dir, task, num_classes):
    config_dir = Path(config_dir)
    config_dir.mkdir(parents=True, exist_ok=True)
    templates_dir = Path(__file__).parent.parent / "faster_autoaugment" / "templates"
    dataset_file = templates_dir / task / "dataset.py.tmpl"
    search_config_file = templates_dir / task / "search.yaml.tmpl"
    dataset_file_destination = config_dir / "dataset.py"
    search_file_destination = config_dir / "search.yaml"

    with search_config_file.open() as f:
        config = AutoAlbumentTemplate(f.read())

    with search_file_destination.open("w") as f:
        f.write(config.substitute(num_classes=num_classes))

    copyfile(dataset_file, dataset_file_destination)

    click.echo()

    click.echo(
        f"Files dataset.py and search.yaml are created in {config_dir}.\n\n"
        f"Next steps:\n"
        f"1. Add the required implementation for dataset methods in "
        + click.style(str(dataset_file_destination), bold=True)
        + "\n"
        + "2. [Optional] Adjust search parameters in "
        + click.style(str(search_file_destination), bold=True)
        + "\n"
        + "3. Run AutoAlbument search with the following command:\n\n"
        + click.style(f"autoalbument-search --config-dir {config_dir}\n", bold=True)
    )


if __name__ == "__main__":
    main()



import importlib.util
from collections import defaultdict

import torch

from autoalbument.utils.hydra import get_dataset_filepath

MAX_VALUES_BY_INPUT_DTYPE = {
    "uint8": 255,
    "float32": 1.0,
}


def get_dataset_cls(dataset_file, dataset_cls_name="SearchDataset"):
    dataset_filepath = get_dataset_filepath(dataset_file)
    spec = importlib.util.spec_from_file_location("dataset", dataset_filepath)
    dataset = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(dataset)
    return getattr(dataset, dataset_cls_name)


class MetricTracker:
    def __init__(self):
        self.reset()

    def reset(self):
        self.metrics = defaultdict(lambda: {"value": 0, "batch_count": 0, "avg": 0})

    def update(self, batch_metrics):
        for name, batch_metric in batch_metrics.items():
            value = batch_metric.item() if isinstance(batch_metric, torch.Tensor) else batch_metric
            self.metrics[name]["value"] += value
            self.metrics[name]["batch_count"] += 1
            self.metrics[name]["avg"] = self.metrics[name]["value"] / self.metrics[name]["batch_count"]

    def get_avg_values(self):
        return [(name, value["avg"]) for name, value in self.metrics.items()]

    def __repr__(self):
        return ", ".join(f"{name}={avg_value:.6f}" for name, avg_value in self.get_avg_values())


def target_requires_grad(target):
    return target == "image_batch"



from enum import Enum
from functools import wraps

import cv2
import torch

MAX_VALUES_BY_DTYPE = {torch.float32: 1.0, torch.float64: 1.0}


class TorchPadding(str, Enum):
    REFLECTION = "reflection"
    BORDER = "border"
    ZEROS = "zeros"


TORCH_TO_OPENCV_BORDER = {
    TorchPadding.REFLECTION: cv2.BORDER_REFLECT,
    TorchPadding.BORDER: cv2.BORDER_REPLICATE,
    TorchPadding.ZEROS: cv2.BORDER_CONSTANT,
}


def clip(img, dtype, maxval):
    return torch.clamp(img, 0, maxval).type(dtype)


def clipped(func):
    @wraps(func)
    def wrapped_function(img, *args, **kwargs):
        dtype = img.dtype
        maxval = MAX_VALUES_BY_DTYPE.get(dtype, 1.0)
        return clip(func(img, *args, **kwargs), dtype, maxval)

    return wrapped_function



import os
import shutil

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf

from autoalbument.config.faster_autoaugment import FasterAutoAugmentSearchConfig
from autoalbument.faster_autoaugment.search import get_faa_seacher
from autoalbument.utils.hydra import get_hydra_config_dir, get_dataset_filepath

OmegaConf.register_resolver("config_dir", get_hydra_config_dir)
cs = ConfigStore.instance()
cs.store(name="search", node=FasterAutoAugmentSearchConfig)


@hydra.main(config_name="search")
def main(cfg):
    print(OmegaConf.to_yaml(cfg, resolve=True))
    cwd = os.getcwd()
    print(f"Working directory : {cwd}")
    dataset_filepath = get_dataset_filepath(cfg.data.dataset_file)
    shutil.copy2(dataset_filepath, cwd)
    faa_searcher = get_faa_seacher(cfg)
    faa_searcher.search()



import os


def get_hydra_config_dir():
    from hydra.core.global_hydra import GlobalHydra

    load_history = GlobalHydra.instance().config_loader().get_load_history()
    cmd_load_traces = [trace for trace in load_history if trace.provider == "command-line"]
    if not cmd_load_traces:
        cmd_load_traces = [trace for trace in load_history if trace.provider == "main"]
    if not cmd_load_traces:
        raise RuntimeError("Couldn't deduce config_dir from load_history.")
    return cmd_load_traces[-1].path.replace("file://", "")


def get_dataset_filepath(dataset_file):
    if os.path.isabs(dataset_file):
        return dataset_file
    base_path = get_hydra_config_dir()
    return os.path.join(base_path, dataset_file)



"""
Based on the official implementation of Faster AutoAugment
https://github.com/moskomule/dda/blob/master/faster_autoaugment/search.py
"""

import copy
import logging
import os
from pathlib import Path
from typing import Tuple

import albumentations as A
import timm
import torch
from albumentations.pytorch import ToTensorV2
from hydra.utils import instantiate
from torch import Tensor, nn
from torch.nn import Flatten
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from autoalbument.faster_autoaugment.metrics import get_average_parameter_change
from autoalbument.faster_autoaugment.utils import MAX_VALUES_BY_INPUT_DTYPE, get_dataset_cls, MetricTracker
from autoalbument.utils.files import symlink
from autoalbument.faster_autoaugment.policy import Policy
import segmentation_models_pytorch as smp

log = logging.getLogger(__name__)


class ClassificationDiscriminator(nn.Module):
    def __init__(self, architecture, pretrained, num_classes):
        super().__init__()
        self.base_model = timm.create_model(architecture, pretrained=pretrained)
        self.base_model.reset_classifier(num_classes)
        self.classifier = self.base_model.get_classifier()
        num_features = self.classifier.in_features
        self.discriminator = nn.Sequential(
            nn.Linear(num_features, num_features), nn.ReLU(), nn.Linear(num_features, 1)
        )

    def forward(self, input: Tensor) -> Tuple[Tensor, Tensor]:
        x = self.base_model.forward_features(input)
        x = self.base_model.global_pool(x).flatten(1)
        return self.classifier(x), self.discriminator(x).view(-1)


class SegmentationDiscriminator(nn.Module):
    def __init__(self, architecture, encoder_architecture, num_classes, pretrained):
        super().__init__()
        model = getattr(smp, architecture)

        self.base_model = model(
            encoder_architecture, encoder_weights=self._get_encoder_weights(pretrained), classes=num_classes
        )
        num_features = self.base_model.encoder.out_channels[-1]
        self.base_model.classification_head = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            Flatten(),
            nn.Linear(num_features, num_features),
            nn.ReLU(),
            nn.Linear(num_features, 1),
        )

    @staticmethod
    def _get_encoder_weights(pretrained):
        if isinstance(pretrained, bool):
            return "imagenet" if pretrained else None
        return pretrained

    def forward(self, input: Tensor) -> Tuple[Tensor, Tensor]:
        mask, discriminator_output = self.base_model(input)
        return mask, discriminator_output.view(-1)


class FasterAutoAugmentBase:
    def __init__(self, cfg):
        self.cfg = cfg
        torch.backends.cudnn.benchmark = self.cfg.cudnn_benchmark
        self.start_epoch = 1
        self.dataloader = self.create_dataloader()
        self.models = self.create_models()
        self.optimizers = self.create_optimizers()
        self.loss = self.create_loss()
        self.metric_tracker = self.create_metric_tracker()
        self.paths = self.create_directories()
        self.tensorboard_writer = self.create_tensorboard_writer()
        self.epoch = None
        if self.cfg.checkpoint_path:
            self.load_checkpoint()
        self.saved_policy_state_dict = self.get_policy_state_dict()

    def create_tensorboard_writer(self):
        if self.cfg.tensorboard_logs_dir:
            return SummaryWriter(os.path.join(self.cfg.tensorboard_logs_dir, os.getcwd().replace(os.sep, ".")))
        return None

    def get_policy_state_dict(self):
        state_dict = copy.deepcopy(self.models["policy"].state_dict())
        for k, v in state_dict.items():
            state_dict[k] = v.to("cpu")
        return state_dict

    def create_directories(self):
        policy_dir = Path.cwd() / "policy"
        policy_dir.mkdir(exist_ok=True)
        if self.cfg.save_checkpoints:
            checkpoints_dir = Path.cwd() / "checkpoints"
            checkpoints_dir.mkdir(exist_ok=True)
        else:
            checkpoints_dir = None
        return {
            "policy_dir": policy_dir,
            "checkpoints_dir": checkpoints_dir,
            "latest_policy_path": policy_dir / "latest.json",
            "latest_checkpoint_path": checkpoints_dir / "latest.pth" if checkpoints_dir is not None else None,
        }

    def get_latest_policy_save_path(self):
        return

    def create_metric_tracker(self):
        return MetricTracker()

    def get_preprocessing_transforms(self):
        preprocessing_config = self.cfg.data.preprocessing
        preprocessing_transforms = []
        if preprocessing_config:
            for preprocessing_transform in preprocessing_config:
                for transform_name, transform_args in preprocessing_transform.items():
                    transform = A.from_dict(
                        {
                            "transform": {
                                "__class_fullname__": "albumentations.augmentations.transforms." + transform_name,
                                **transform_args,
                            }
                        }
                    )
                    preprocessing_transforms.append(transform)
        return preprocessing_transforms

    def create_preprocessing_transform(self):
        normalization_config = self.cfg.data.normalization
        input_dtype = self.cfg.data.input_dtype
        transform = A.Compose(
            [
                *self.get_preprocessing_transforms(),
                A.Normalize(
                    mean=normalization_config.mean,
                    std=normalization_config.std,
                    max_pixel_value=MAX_VALUES_BY_INPUT_DTYPE[input_dtype],
                ),
                ToTensorV2(transpose_mask=True),
            ]
        )
        log.info(f"Preprocessing transform:\n{transform}")
        return transform

    def create_dataloader(self):
        dataset_cls = get_dataset_cls(self.cfg.data.dataset_file)
        transform = self.create_preprocessing_transform()
        dataset = dataset_cls(transform=transform)
        dataloader = instantiate(self.cfg.data.dataloader, dataset=dataset)
        return dataloader

    def create_optimizers(self):
        optimizer_config = self.cfg.optim
        main_optimizer = instantiate(optimizer_config.main, params=self.models["main"].parameters())
        policy_optimizer = instantiate(optimizer_config.policy, params=self.models["policy"].parameters())
        return {
            "main": main_optimizer,
            "policy": policy_optimizer,
        }

    def create_main_model(self):
        raise NotImplementedError

    def create_policy_model(self):
        policy_model_cfg = self.cfg.policy_model
        normalization_cfg = self.cfg.data.normalization
        policy_model = Policy.faster_auto_augment_policy(
            policy_model_cfg.num_sub_policies,
            policy_model_cfg.temperature,
            policy_model_cfg.operation_count,
            policy_model_cfg.num_chunks,
            mean=torch.tensor(normalization_cfg.mean),
            std=torch.tensor(normalization_cfg.std),
        ).to(self.cfg.device)
        return policy_model

    def create_models(self):
        main_model = self.create_main_model()
        policy_model = self.create_policy_model()
        return {"main": main_model, "policy": policy_model}

    def policy_forward_for_policy_train(self, a_input, a_target):
        raise NotImplementedError

    def gradient_penalty(self, real: Tensor, fake: Tensor) -> Tensor:
        alpha = real.new_empty(real.size(0), 1, 1, 1).uniform_(0, 1)
        interpolated = alpha * real + (1 - alpha) * fake
        interpolated.requires_grad_()
        _, output = self.models["main"](interpolated)
        grad = torch.autograd.grad(
            outputs=output,
            inputs=interpolated,
            grad_outputs=torch.ones_like(output),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]
        return (grad.norm(2, dim=1) - 1).pow(2).mean()

    def create_loss(self):
        return nn.CrossEntropyLoss().to(self.cfg.device)

    def wgan_loss(
        self, n_input: Tensor, n_target: Tensor, a_input: Tensor, a_target: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        ones = n_input.new_tensor(1.0)
        self.models["main"].requires_grad_(True)
        self.models["main"].zero_grad()
        output, n_output = self.models["main"](n_input)
        loss = self.cfg.policy_model.task_factor * self.loss(output, n_target)
        loss.backward(retain_graph=True)
        d_n_loss = n_output.mean()
        d_n_loss.backward(-ones)

        with torch.no_grad():
            a_input = self.models["policy"].denormalize_(a_input)
            augmented = self.models["policy"]({"image_batch": a_input})["image_batch"]

        _, a_output = self.models["main"](augmented)
        d_a_loss = a_output.mean()
        d_a_loss.backward(ones)
        gp = self.cfg.policy_model.gp_factor * self.gradient_penalty(n_input, augmented)
        gp.backward()
        self.optimizers["main"].step()

        self.models["main"].requires_grad_(False)
        self.models["policy"].zero_grad()

        augmented_input, maybe_augmented_target = self.policy_forward_for_policy_train(a_input, a_target)

        _output, a_output = self.models["main"](augmented_input)
        _loss = self.cfg.policy_model.task_factor * self.loss(_output, maybe_augmented_target)
        _loss.backward(retain_graph=True)
        a_loss = a_output.mean()
        a_loss.backward(-ones)
        self.optimizers["policy"].step()
        return loss + _loss, -d_n_loss + d_a_loss + gp, -a_loss

    def train_step(self, input, target):
        b = input.size(0) // 2
        a_input, a_target = input[:b], target[:b]
        n_input, n_target = input[b:], target[b:]
        loss, d_loss, a_loss = self.wgan_loss(n_input, n_target, a_input, a_target)
        return {
            "loss": loss,
            "d_loss": d_loss,
            "a_loss": a_loss,
        }

    def get_progress_bar_description(self):
        return f"Epoch: {self.epoch}. {self.metric_tracker}"

    def train_epoch(self):
        self.metric_tracker.reset()
        pbar = tqdm(total=len(self.dataloader))
        for input, target in self.dataloader:
            input = input.to(self.cfg.device)
            target = target.to(self.cfg.device)
            loss_dict = self.train_step(input, target)
            self.metric_tracker.update(loss_dict)
            pbar.set_description(self.get_progress_bar_description())
            pbar.update()

        average_parameter_change = get_average_parameter_change(
            self.saved_policy_state_dict, self.get_policy_state_dict()
        )
        self.saved_policy_state_dict = self.get_policy_state_dict()
        self.metric_tracker.update({"Average Parameter change": average_parameter_change})
        pbar.set_description(self.get_progress_bar_description())
        pbar.close()

    def write_tensorboard_logs(self):
        if not self.tensorboard_writer:
            return
        for metric, avg_value in self.metric_tracker.get_avg_values():
            self.tensorboard_writer.add_scalar(metric, avg_value, self.epoch)

    def save_policy(self):

        transform = self.models["policy"].create_transform(
            input_dtype=self.cfg.data.input_dtype,
            preprocessing_transforms=self.get_preprocessing_transforms(),
        )
        policy_save_path = self.paths["policy_dir"] / f"epoch_{self.epoch}.json"
        A.save(transform, str(policy_save_path))
        symlink(policy_save_path, self.paths["latest_policy_path"])
        log.info(
            f"Policy is saved to {policy_save_path}. "
            f"{self.paths['latest_policy_path']} now also points to this policy file."
        )

    def save_checkpoint(self):
        checkpoint = {
            "epoch": self.epoch,
            "models": {k: v.state_dict() for k, v in self.models.items()},
            "optimizers": {k: v.state_dict() for k, v in self.optimizers.items()},
        }
        checkpoint_save_path = self.paths["checkpoints_dir"] / f"epoch_{self.epoch}.pth"
        torch.save(checkpoint, checkpoint_save_path)
        symlink(checkpoint_save_path, self.paths["latest_checkpoint_path"])
        log.info(
            f"Checkpoint is saved to {checkpoint_save_path}. "
            f"{self.paths['latest_checkpoint_path']} now also points to this checkpoint file."
        )

    def load_checkpoint(self):
        checkpoint_path = self.cfg.checkpoint_path
        log.info(f"Loading checkpoint {checkpoint_path}")
        checkpoint = torch.load(checkpoint_path, map_location=self.cfg.device)
        for name, model in self.models.items():
            model.load_state_dict(checkpoint["models"][name])
        for name, optimizer in self.optimizers.items():
            optimizer.load_state_dict(checkpoint["optimizers"][name])
        self.start_epoch = checkpoint["epoch"] + 1
        self.saved_policy_state_dict = self.get_policy_state_dict()

    def save(self):
        self.save_policy()
        if self.cfg.save_checkpoints:
            self.save_checkpoint()

    def get_search_summary(self):
        summary = [
            f"\n\nSearch is finished.\n"
            f"- Configuration files for policies found at each epoch are located in {self.paths['policy_dir']}."
        ]
        if self.cfg.save_checkpoints:
            summary.append(
                f"- Checkpoint files for models and optimizers at each epoch are located in "
                f"{self.paths['checkpoints_dir']}."
            )

        latest_policy_path = self.paths["latest_policy_path"]
        load_command = f'transform = A.load("{latest_policy_path}")\n'

        separator = "-" * len(load_command)
        if self.cfg.task == "classification":
            transform_text = "transformed = transform(image=image)\n" 'transformed_image = transformed["image"]\n'
        else:
            transform_text = (
                "transformed = transform(image=image, mask=mask)\n"
                'transformed_image = transformed["image"]\n'
                'transformed_mask = transformed["mask"]\n'
            )

        summary.append(
            f"\nUse Albumentations to load the found policies and transform images:\n"
            f"{separator}\n"
            f"import albumentations as A\n\n"
            f'transform = A.load("{latest_policy_path}")\n' + transform_text + f"{separator}\n"
        )

        return "\n".join(summary)

    def search(self):
        for epoch in range(self.start_epoch, self.cfg.optim.epochs + 1):
            self.epoch = epoch
            self.train_epoch()
            self.write_tensorboard_logs()
            self.save()

        log.info(self.get_search_summary())


class FAAClassification(FasterAutoAugmentBase):
    def create_main_model(self):
        model_cfg = self.cfg.classification_model
        main_model = ClassificationDiscriminator(
            model_cfg.architecture, num_classes=model_cfg.num_classes, pretrained=model_cfg.pretrained
        ).to(self.cfg.device)
        return main_model

    def create_loss(self):
        return nn.CrossEntropyLoss().to(self.cfg.device)

    def policy_forward_for_policy_train(self, a_input, a_target):
        output = self.models["policy"]({"image_batch": a_input})["image_batch"]
        return output, a_target


class FAASemanticSegmentation(FasterAutoAugmentBase):
    def create_main_model(self):
        model_cfg = self.cfg.semantic_segmentation_model
        main_model = SegmentationDiscriminator(
            model_cfg.architecture,
            encoder_architecture=model_cfg.encoder_architecture,
            num_classes=model_cfg.num_classes,
            pretrained=model_cfg.pretrained,
        ).to(self.cfg.device)
        return main_model

    def create_loss(self):
        return nn.BCEWithLogitsLoss().to(self.cfg.device)

    def policy_forward_for_policy_train(self, a_input, a_target):
        output = self.models["policy"]({"image_batch": a_input, "mask_batch": a_target})
        return output["image_batch"], output["mask_batch"]


def get_faa_seacher(cfg):
    task = cfg.task
    if task == "semantic_segmentation":
        return FAASemanticSegmentation(cfg)
    elif task == "classification":
        return FAAClassification(cfg)
    raise ValueError(f"Unsupported task: {task}. Supported tasks: classification, semantic_segmentation.")



import random

import torch

from autoalbument.faster_autoaugment.albumentations_pytorch.affine import (
    get_scaling_matrix,
    get_rotation_matrix,
    warp_affine,
)
from autoalbument.faster_autoaugment.albumentations_pytorch.utils import MAX_VALUES_BY_DTYPE, TorchPadding, clipped


def solarize(img_batch, threshold):
    dtype = img_batch.dtype
    max_val = MAX_VALUES_BY_DTYPE[dtype]
    return torch.where(img_batch >= threshold, max_val - img_batch, img_batch)


@clipped
def shift_rgb(img_batch, r_shift, g_shift, b_shift):
    result_img_batch = img_batch.clone()
    shifts = [r_shift, g_shift, b_shift]
    for i, shift in enumerate(shifts):
        result_img_batch[:, i] = result_img_batch[:, i] + shift
    return result_img_batch


@clipped
def brightness_adjust(img_batch, beta):
    return img_batch + beta


@clipped
def contrast_adjust(img_batch, alpha):
    return img_batch * alpha


def vflip(img_batch):
    return torch.flip(img_batch, [-2]).contiguous()


def hflip(img_batch):
    return torch.flip(img_batch, [-1]).contiguous()


def shift_x(img_batch, dx, padding_mode=TorchPadding.REFLECTION):
    scaling_matrix = get_scaling_matrix(img_batch, dx=dx)
    return warp_affine(img_batch, scaling_matrix, padding_mode)


def shift_y(img_batch, dy, padding_mode=TorchPadding.REFLECTION):
    scaling_matrix = get_scaling_matrix(img_batch, dy=dy)
    return warp_affine(img_batch, scaling_matrix, padding_mode)


def rotate(img_batch, angle, padding_mode=TorchPadding.REFLECTION):
    rotation_matrix = get_rotation_matrix(img_batch, angle=angle)
    return warp_affine(img_batch, rotation_matrix, padding_mode)


def scale(img_batch, scale, padding_mode=TorchPadding.REFLECTION):
    rotation_matrix = get_rotation_matrix(img_batch, scale=scale)
    return warp_affine(img_batch, rotation_matrix, padding_mode)


def cutout(img_batch, num_holes, hole_size, fill_value=0):
    img_batch = img_batch.clone()
    height, width = img_batch.shape[-2:]
    for i in range(len(img_batch)):
        for _n in range(num_holes):
            y1 = random.randint(0, height - hole_size)
            x1 = random.randint(0, width - hole_size)
            y2 = y1 + hole_size
            x2 = x1 + hole_size
            img_batch[i, :, y1:y2, x1:x2] = fill_value
    return img_batch



"""
Based on the official implementation of Faster AutoAugment
https://github.com/moskomule/dda/blob/master/dda/operations.py
"""

import warnings

import albumentations as A
import torch
from torch import nn
from torch.autograd import Function
from torch.distributions import RelaxedBernoulli

from autoalbument.faster_autoaugment.albumentations_pytorch import functional as F
from autoalbument.faster_autoaugment.utils import target_requires_grad


class _STE(Function):
    @staticmethod
    def forward(ctx, input_forward, input_backward):
        ctx.shape = input_backward.shape
        return input_forward

    @staticmethod
    def backward(ctx, grad_in):
        return None, grad_in.sum_to_size(ctx.shape)


def ste(input_forward: torch.Tensor, input_backward: torch.Tensor):
    return _STE.apply(input_forward, input_backward).clone()


class Operation(nn.Module):
    def __init__(
        self,
        temperature: float = 0.1,
        value_range=(0.0, 1.0),
        has_magnitude=True,
        is_spatial_level=False,
        ste=False,
        requires_uint8_scaling=False,
    ):

        super().__init__()
        self.is_spatial_level = is_spatial_level
        self.ste = ste
        self.value_range = value_range
        self.requires_uint8_scaling = requires_uint8_scaling
        self._magnitude = nn.Parameter(torch.empty(1)) if has_magnitude else None
        self._probability = nn.Parameter(torch.empty(1))
        self.register_buffer("temperature", torch.empty(1).fill_(temperature))

    def forward(self, input):
        mask = self.get_probability_mask(input["image_batch"].size(0))
        operation_output = self.operation(input)

        targets = input.keys()
        output = {}
        for target in targets:
            with torch.set_grad_enabled(target_requires_grad(target)):
                output[target] = (mask * operation_output[target] + (1 - mask) * input[target]).clamp_(0, 1)

        return output

    def get_probability_mask(self, batch_size):
        size = (batch_size, 1, 1)
        return RelaxedBernoulli(self.temperature, self.probability).rsample(size)

    @property
    def magnitude(self):
        if self._magnitude is None:
            return None
        return self._magnitude.clamp(0.0, 1.0)

    @property
    def targets(self):
        return {
            "image_batch": self.apply_operation,
            "mask_batch": self.apply_operation_to_mask,
        }

    @property
    def probability(self):
        return self._probability.clamp(0.0, 1.0)

    def __repr__(self):
        probability = self.probability.item()
        magnitude = self.magnitude
        if magnitude is not None:
            magnitude = magnitude.item()
        temperature = self.temperature.item()  # type: ignore

        repr_str = self.__class__.__name__
        repr_str += f"(probability={probability:.3f}, "
        if magnitude is not None:
            repr_str += f"magnitude={magnitude:.3f}, "
        repr_str += f"temperature={temperature:.3f})"
        return repr_str

    def apply_operation_to_mask(self, input, value):
        with torch.no_grad():
            if self.is_spatial_level:
                return self.apply_operation(input, value)
            return input

    def operation(self, input):
        magnitude = self.magnitude
        value = convert_value_range(magnitude, self.value_range) if magnitude is not None else None

        output = {}
        targets = input.keys()
        for target in targets:
            operation_fn = self.targets[target]
            operation_output = operation_fn(input[target], value)
            if target_requires_grad(target) and self.ste:
                operation_output = ste(operation_output, magnitude)
            output[target] = operation_output
        return output

    def create_transform(self, input_dtype, p):
        magnitude = self.magnitude
        if magnitude is not None:
            value = convert_value_range(magnitude, self.value_range)
            transform_param = value_to_transform_param(
                value, input_dtype, requires_uint8_scaling=self.requires_uint8_scaling
            )
        else:
            transform_param = None
        return self.as_transform(transform_param, p)


def value_to_transform_param(value, input_dtype="float32", requires_uint8_scaling=False):
    value = value.item()
    if input_dtype == "uint8" and requires_uint8_scaling:
        value = int(value * 255.0)
    return value


def convert_value_range(value, new_range):
    lower, upper = new_range
    total = upper - lower
    return value * total + lower


class ShiftRGB(Operation):
    def __init__(self, temperature, shift_r=False, shift_g=False, shift_b=False):
        super().__init__(temperature, value_range=(-1.0, 1.0), requires_uint8_scaling=True)
        self.shift_r = shift_r
        self.shift_g = shift_g
        self.shift_b = shift_b

    def apply_operation(self, input, value):
        return F.shift_rgb(
            input,
            r_shift=value if self.shift_r else 0.0,
            g_shift=value if self.shift_g else 0.0,
            b_shift=value if self.shift_b else 0.0,
        )

    def as_transform(self, value, p):
        limit = (value, value)
        return A.RGBShift(
            r_shift_limit=limit if self.shift_r else 0.0,
            g_shift_limit=limit if self.shift_g else 0.0,
            b_shift_limit=limit if self.shift_b else 0.0,
            p=p,
        )


class RandomBrightness(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(-1.0, 1.0))

    def apply_operation(self, input, value):
        return F.brightness_adjust(input, beta=value)

    def as_transform(self, value, p):
        return A.RandomBrightnessContrast(brightness_limit=(value, value), contrast_limit=0, p=p)


class RandomContrast(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(0.0, 10.0))

    def apply_operation(self, input, value):
        return F.contrast_adjust(input, alpha=value)

    def as_transform(self, value, p):
        return A.RandomBrightnessContrast(brightness_limit=0, contrast_limit=(value, value), p=p)


class Solarize(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, requires_uint8_scaling=True)

    def apply_operation(self, input, value):
        return F.solarize(input, threshold=value)

    def as_transform(self, value, p):
        return A.Solarize(threshold=value, p=p)


class HorizontalFlip(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, has_magnitude=False, is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.hflip(input)

    def as_transform(self, value, p):
        return A.HorizontalFlip(p=p)


class VerticalFlip(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, has_magnitude=False, is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.vflip(input)

    def as_transform(self, value, p):
        return A.VerticalFlip(p=p)


class ShiftX(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(-1.0, 1.0), is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.shift_x(input, dx=value)

    def as_transform(self, value, p):
        return A.ShiftScaleRotate(
            shift_limit_x=(value, value),
            shift_limit_y=(0, 0),
            rotate_limit=(0, 0),
            scale_limit=(0, 0),
            p=p,
        )


class ShiftY(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(-1.0, 1.0), is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.shift_y(input, dy=value)

    def as_transform(self, value, p):
        return A.ShiftScaleRotate(
            shift_limit_x=(0, 0),
            shift_limit_y=(value, value),
            rotate_limit=(0, 0),
            scale_limit=(0, 0),
            p=p,
        )


class Scale(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(0 + 1e-8, 10.0), is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.scale(input, scale=value)

    def as_transform(self, value, p):
        return A.ShiftScaleRotate(
            shift_limit_x=(0, 0),
            shift_limit_y=(0, 0),
            rotate_limit=(0, 0),
            scale_limit=(value, value),
            p=p,
        )


class Rotate(Operation):
    def __init__(self, temperature):
        super().__init__(temperature, value_range=(-180, 180), is_spatial_level=True)

    def apply_operation(self, input, value):
        return F.rotate(input, angle=value)

    def as_transform(self, value, p):
        return A.ShiftScaleRotate(
            shift_limit_x=(0, 0),
            shift_limit_y=(0, 0),
            rotate_limit=(value, value),
            scale_limit=(0, 0),
            p=p,
        )


class Cutout(Operation):
    def __init__(self, temperature, value_range=(0.0, 1.0)):
        super().__init__(temperature, value_range=value_range)
        self.register_buffer("saved_image_shape", torch.Tensor([0, 0]).type(torch.int64))
        self.is_image_shape_saved = False

    def _save_image_shape(self, image_shape):
        if not torch.equal(self.saved_image_shape, image_shape):
            if self.is_image_shape_saved:
                warnings.warn(
                    f"Shape of images in a batch changed between iterations "
                    f"from {self.saved_image_shape} to {image_shape}. "
                    f"This will affect the created Albumentations transform. "
                    f"The transform will use the shape {image_shape} to initialize its parameters",
                    RuntimeWarning,
                )
            self.is_image_shape_saved = True
            self.saved_image_shape = image_shape

    def apply_operation(self, input, value):
        image_shape = input.shape[-2:]
        self._save_image_shape(torch.tensor(image_shape).to(input.device))
        return self._apply_cutout(input, value, image_shape)

    def as_transform(self, value, p):
        image_shape = self.saved_image_shape
        return self._as_cutout_transform(value, p, image_shape)

    def _apply_cutout(self, input, value, image_shape):
        raise NotImplementedError

    def _as_cutout_transform(self, value, p, image_shape):
        raise NotImplementedError


class CutoutFixedNumerOfHoles(Cutout):
    def __init__(self, temperature, num_holes=16):
        super().__init__(temperature)
        self.num_holes = num_holes

    def _calculate_hole_size(self, value, image_shape):
        height, width = image_shape
        min_size = min(height, width)
        return max(int(min_size * value), 1)

    def _apply_cutout(self, input, value, image_shape):
        hole_size = self._calculate_hole_size(value, image_shape)
        return F.cutout(input, num_holes=self.num_holes, hole_size=hole_size)

    def _as_cutout_transform(self, value, p, image_shape):
        hole_size = self._calculate_hole_size(value, image_shape)
        return A.CoarseDropout(
            min_holes=self.num_holes,
            max_holes=self.num_holes,
            max_height=hole_size,
            max_width=hole_size,
            p=p,
        )


class CutoutFixedSize(Cutout):
    def __init__(self, temperature, hole_size_divider=16, max_holes=16):
        super().__init__(temperature, value_range=(0, max_holes))
        self.hole_size_divider = hole_size_divider

    def _calculate_hole_size(self, value, image_shape):
        height, width = image_shape
        min_size = min(height, width)
        hole_size = int(min_size // self.hole_size_divider)
        return max(hole_size, 1)

    def _apply_cutout(self, input, value, image_shape):
        hole_size = self._calculate_hole_size(value, image_shape)
        return F.cutout(input, num_holes=int(value), hole_size=hole_size)

    def _as_cutout_transform(self, value, p, image_shape):
        hole_size = self._calculate_hole_size(value, image_shape)
        num_holes = max(int(value), 1)
        return A.CoarseDropout(
            min_holes=num_holes,
            max_holes=num_holes,
            max_height=hole_size,
            max_width=hole_size,
            p=p,
        )



"""
Based on the official implementation of Faster AutoAugment
https://github.com/moskomule/dda/blob/master/faster_autoaugment/policy.py
"""

import random
from copy import deepcopy
from typing import Dict

import albumentations as A
import torch
from albumentations.core.composition import Compose, OneOf
from albumentations.pytorch import ToTensorV2
from torch import Tensor, nn

from autoalbument.faster_autoaugment.operations import (
    CutoutFixedNumerOfHoles,
    CutoutFixedSize,
    HorizontalFlip,
    RandomBrightness,
    RandomContrast,
    Rotate,
    Scale,
    ShiftRGB,
    ShiftX,
    ShiftY,
    Solarize,
    VerticalFlip,
)
from autoalbument.faster_autoaugment.utils import MAX_VALUES_BY_INPUT_DTYPE, target_requires_grad


class SubPolicyStage(nn.Module):
    def __init__(
        self,
        operations: nn.ModuleList,
        temperature: float,
    ):
        super().__init__()
        self.operations = operations
        self._weights = nn.Parameter(torch.ones(len(self.operations)))
        self.temperature = temperature

    def forward(self, input: Dict[str, Tensor]) -> Dict[str, Tensor]:
        targets = input.keys()
        operation_outputs = [op(input) for op in self.operations]
        output = {}
        for target in targets:
            with torch.set_grad_enabled(target_requires_grad(target)):
                output[target] = (
                    torch.stack([op[target] for op in operation_outputs]) * self.weights.view(-1, 1, 1, 1, 1)
                ).sum(0)
        return output

    @property
    def weights(self):
        return self._weights.div(self.temperature).softmax(0)

    def create_transform(self, input_dtype):
        weights = self.weights.detach().cpu().numpy().tolist()
        probabilities = [op.probability.item() for op in self.operations]
        true_probabilities = [w * p for (w, p) in zip(weights, probabilities)]
        assert sum(true_probabilities) <= 1.0
        transforms = []
        p_sum = 0
        for operation, p in zip(self.operations, true_probabilities):
            transforms.append(operation.create_transform(input_dtype, p))
            p_sum += p
        transforms.append(A.NoOp(p=1.0 - p_sum))
        return OneOf(transforms, p=1)


class SubPolicy(nn.Module):
    def __init__(
        self,
        sub_policy_stage: SubPolicyStage,
        operation_count: int,
    ):
        super().__init__()
        self.stages = nn.ModuleList([deepcopy(sub_policy_stage) for _ in range(operation_count)])

    def forward(self, input: Dict[str, Tensor]) -> Dict[str, Tensor]:
        for stage in self.stages:
            input = stage(input)
        return input

    def create_transform(self, input_dtype, p):
        return A.Sequential([stage.create_transform(input_dtype) for stage in self.stages], p=p)


class Policy(nn.Module):
    def __init__(
        self,
        operations: nn.ModuleList,
        num_sub_policies: int,
        temperature: float,
        operation_count: int,
        num_chunks: int,
        mean: Tensor,
        std: Tensor,
    ):
        super().__init__()
        self.sub_policies = nn.ModuleList(
            [SubPolicy(SubPolicyStage(operations, temperature), operation_count) for _ in range(num_sub_policies)]
        )
        self.num_sub_policies = num_sub_policies
        self.temperature = temperature
        self.operation_count = operation_count
        self.num_chunks = num_chunks
        self.register_buffer("_mean", mean)
        self.register_buffer("_std", std)

        for p in self.parameters():
            nn.init.uniform_(p, 0, 1)

    def forward(self, input):
        targets = input.keys()

        chunked_input = {}
        num_chunks = None
        for target in targets:
            chunked_input[target] = input[target].chunk(self.num_chunks)
            num_chunks = len(chunked_input[target])

        prepared_chunked_input = []
        for i in range(num_chunks):
            prepared_chunked_input.append({target: chunked_input[target][i] for target in targets})

        if self.num_chunks > 1:
            output = {}
            out_chunks = [self._forward(inp) for inp in prepared_chunked_input]
            for target in targets:
                output[target] = torch.cat([out_chunk[target] for out_chunk in out_chunks], dim=0)
        else:
            output = self._forward(input)

        output["image_batch"] = self.normalize_(output["image_batch"])
        return output

    def _forward(self, input: Dict[str, Tensor]) -> Dict[str, Tensor]:
        index = random.randrange(self.num_sub_policies)
        return self.sub_policies[index](input)

    def normalize_(self, input):
        # [0, 1] -> [-1, 1]
        return input.add_(-self._mean[:, None, None]).div_(self._std[:, None, None])

    def denormalize_(self, input):
        # [-1, 1] -> [0, 1]
        return input.mul_(self._std[:, None, None]).add_(self._mean[:, None, None])

    @staticmethod
    def dda_operations(temperature):
        return [
            ShiftRGB(shift_r=True, temperature=temperature),
            ShiftRGB(shift_g=True, temperature=temperature),
            ShiftRGB(shift_b=True, temperature=temperature),
            RandomBrightness(temperature=temperature),
            RandomContrast(temperature=temperature),
            Solarize(temperature=temperature),
            HorizontalFlip(temperature=temperature),
            VerticalFlip(temperature=temperature),
            Rotate(temperature=temperature),
            ShiftX(temperature=temperature),
            ShiftY(temperature=temperature),
            Scale(temperature=temperature),
            CutoutFixedNumerOfHoles(temperature=temperature),
            CutoutFixedSize(temperature=temperature),
        ]

    @staticmethod
    def faster_auto_augment_policy(
        num_sub_policies: int,
        temperature: float,
        operation_count: int,
        num_chunks: int,
        mean: torch.Tensor,
        std: torch.Tensor,
    ):
        return Policy(
            nn.ModuleList(Policy.dda_operations(temperature)),
            num_sub_policies,
            temperature,
            operation_count,
            num_chunks,
            mean=mean,
            std=std,
        )

    def create_transform(self, input_dtype="float32", preprocessing_transforms=()):
        sub_policy_p = 1 / len(self.sub_policies)
        return Compose(
            [
                *preprocessing_transforms,
                OneOf([sp.create_transform(input_dtype, p=sub_policy_p) for sp in self.sub_policies], p=1),
                A.Normalize(
                    mean=self._mean.tolist(),
                    std=self._std.tolist(),
                    max_pixel_value=MAX_VALUES_BY_INPUT_DTYPE[input_dtype],
                ),
                ToTensorV2(transpose_mask=True),
            ]
        )
